{ Gestion de Filtres pour le BDE, peut virtuellement remplacer une requte
  SQL sans l'attente de l'excution, le temps tant rparti  chaque browse
  de la base (le BDE saute les records ne correspondant pas au filtre).
  La prsente unit n'est qu'un test avant la mise en place de filtres
  plus complexes. Elle implmente toutefois trois objets filtres fonctionnels
  dont la seule limite est de n'tre constitu que d'une arbre simple  trois
  noeuds: un champ, un oprateur, une valeur.
  La documentation utilise est:
     * Borland Database Engine 2.0
     * Source de l'Interface BDE de DELPHI
     * sans oublier quelques discussions trs fructueuses avec plusieurs
       personnes sur Compuserve, les centaines de Ko de textes et d'exemples
       dont j'ai pu tirer enseignement.
     * Special thanks to Jim Pierce for the 1st version that included
       Long integer, String16 and Float filters.

  Copyright Olivier DAHAN. 25 Avril 1995. dbFilter

  Comment:
          (1) Time Filter
          The BDE stocks Time as the sum of milliseconds since 0:00
          and you can't know what are those msecs... When you
          transpose a time into a TdateTime var, msecs are always 0.
          So, no exact match can be found. When using the Time filter,
          alway use "Greater or Equal" test, never EQUAL that gives nothing...

  Versions History
  V 0.1 Jim Pierce longint, string16, float
  V 1.0 April 25, 1995, DCR for palette icons, date filter
  V 1.5 May 7, 1995, time filter, Bool Filter, SmallInt Filter.
  }



unit Dbfilter;

interface

uses
  CLASSES,
  WinTypes,
  WinProcs,
  SYSUTILS,
  DBIPROCS,
  DbiTypes,
  DB,
  DBTABLES;

type
  String16 = String[16];
  String32 = String[32];
  Char20   = Array[0..19] of Char;

  TCANExp =
    record
      Header       : CANExpr;
      BinaryNode   : CANBinary;
      FieldNode    : CANField;
      ConstNode    : CANConst;
      FieldLiteral,
      ValueLiteral : Char20;
    end;

  TFilter =
    class(TComponent)
      private
        FDataLink   : TFieldDataLink;
        FOperator   : DBITypes.CANOp;
        FHandle     : hDBICur;
        FValueType  : Word;
        FValueSize  : Word;
        CANExp      : TCANExp;
        hFilter     : hDBIFilter;
        function GetDataField: string;
        procedure SetDataField(const Value: string);
        function GetDataSource: TDataSource;
        procedure SetDataSource(Value: TDataSource);
        function GetHandle : hDBICur;
        function GetFieldNum : Word;
      protected
        function GetValueType : Word; virtual; abstract;
        function GetValueSize : Word; virtual; abstract;
        procedure MoveValue; virtual;
      public
        procedure Add;
        procedure Activate;
        procedure Deactivate;
        procedure Drop;
        constructor Create(AOwner: TComponent); override;
        destructor Destroy; override;
        property Handle : hDBICur read GetHandle;
        property FieldNum : Word read GetFieldNum;
        property ValueType : Word read GetValueType;
        property ValueSize : Word read GetValueSize;
      published
        property DataField : String read GetDataField write SetDataField;
        property DataSource : TDataSource read GetDataSource write SetDataSource;
        property Operator : CANOp read FOperator write FOperator;
    end;

  TIntegerFilter =
    class(TFilter)
      private
        FValue : LongInt;
      protected
        constructor Create(AOwner: TComponent); override;
        function GetValueType : Word; override;
        function GetValueSize : Word; override;
        procedure MoveValue; override;
      published
        property Value : LongInt read FValue write FValue;
    end;

  TFloatFilter =
    class(TFilter)
      private
        FValue : Double;
      protected
        constructor Create(AOwner: TComponent); override;
        function GetValueType : Word; override;
        function GetValueSize : Word; override;
        procedure MoveValue; override;
      published
        property Value : Double read FValue write FValue;
    end;

  TStringFilter =
    class(TFilter)
      private
        FValue : String16;
      protected
        constructor Create(AOwner: TComponent); override;
        function GetValueType : Word; override;
        function GetValueSize : Word; override;
        procedure MoveValue; override;
      published
        property Value : String16 read FValue write FValue;
    end;

  TDateFilter =
    class(TFilter)
      private
        FValue : Longint;
        FVD    : TDateTime;
      protected
        constructor Create(AOwner: TComponent); override;
        function GetValueType : Word; override;
        function GetValueSize : Word; override;
        procedure MoveValue; override;
        function GetDate: String16;
        procedure SetDate( dtArg: String16 );
      published
        property Value : String16 read GetDate write SetDate;
    end;

TTimeFilter =
    class(TFilter)
      private
        FValue : Longint;
        FVD    : TDateTime;
      protected
        constructor Create(AOwner: TComponent); override;
        function GetValueType : Word; override;
        function GetValueSize : Word; override;
        procedure MoveValue; override;
        function GetTime: String16;
        procedure SetTime( dtArg: String16 );
      published
        property Value : String16 read GetTime write SetTime;
    end;

TBoolFilter =
    class(TFilter)
      private
        FValue : Bool;
        FVD    : Boolean;
      protected
        constructor Create(AOwner: TComponent); override;
        function GetValueType : Word; override;
        function GetValueSize : Word; override;
        procedure MoveValue; override;
      published
        property Value : Boolean read Fvd write Fvd;
    end;


  TSmallIntFilter =
    class(TFilter)
      private
        FValue : Integer;
      protected
        constructor Create(AOwner: TComponent); override;
        function GetValueType : Word; override;
        function GetValueSize : Word; override;
        procedure MoveValue; override;
      published
        property Value : integer read FValue write FValue;
    end;

  TTimeStampFilter =
    class(TFilter)
      private
        FValue : double;
        FVD    : TDateTime;
      protected
        constructor Create(AOwner: TComponent); override;
        function GetValueType : Word; override;
        function GetValueSize : Word; override;
        procedure MoveValue; override;
        function GetTimeStamp: String32;
        procedure SetTimeStamp( dtArg: String32 );
      published
        property Value : String32 read GetTimeStamp write SetTimeStamp;
    end;

procedure Register;

implementation

constructor TFilter.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FDataLink := TFieldDataLink.Create;
end;

destructor TFilter.Destroy;
begin
  FDataLink.Free;
  FDataLink := nil;
  inherited Destroy;
end;

function TFilter.GetDataField : String;
begin
  Result := FDataLink.FieldName;
end;

procedure TFilter.SetDataField(const Value : String);
begin
  FDataLink.FieldName := Value;
end;

function TFilter.GetDataSource: TDataSource;
begin
  Result := FDataLink.DataSource;
end;

procedure TFilter.SetDataSource(Value: TDataSource);
begin
  FDataLink.DataSource := Value;
end;

function TFilter.GetHandle : hDBICur;
begin
  GetHandle := FDataLink.DataSource.DataSet.Handle;
end;

function TFilter.GetFieldNum : Word;
begin
  GetFieldNum := FDataLink.DataSource.Dataset.FieldByName(DataField).FieldNo;
end;

procedure TFilter.Add;
var
  FieldName : String32;
begin
  FillChar(CANExp, SizeOf(CANExp), 0);
  with CANExp.header do
    begin
      iVer := 1;                     { Version "1" }
      iTotalSize := SizeOf(CANExp);  { Taille totale du filtre }
      iNodes := 3;                   { nombre de noeuds }
      iNodeStart := 10;              { Offset dpart des noeuds expression }
      iLiteralStart := 36;           { Offset dpart des litraux }
    end;
  with CANExp.BinaryNode do
    begin
      NodeClass := nodeBINARY;
      CANOp := Operator;             { Greater than, Less than, Equal etc. }
      iOperand1 := 8;                { Offset du FieldNode }
      iOperand2 := 16;               { Offset du ConstNode }
    end;
  with CANExp.FieldNode do
    begin
      NodeClass := nodeFIELD;
      CANOp := canFIELD2;
      iFieldNum := FieldNum;         { N du champ dans la table }
      iNameOffset := 0;              { szField est le literal  l'offset 0 }
    end;
  with CANExp.ConstNode do
    begin
      NodeClass := nodeCONST;
      CANOp := canCONST2;
      iType := ValueType;
      iSize := ValueSize;            { taille de la constante }
      iOffset := 20;                 { les Literaaux sont separs par 20 octets
                                       taille fige dans cette version (arbitraire) }
    end;
  FieldName := DataField;
  Move(FieldName[1], CANExp.FieldLiteral, Length(FieldName));
  MoveValue;

  DbiAddFilter(Handle, 0, 1, FALSE, Addr(CANExp), nil, hFilter);
end;

procedure TFilter.Activate;
begin
  DbiActivateFilter(Handle, hFilter);
end;

procedure TFilter.Deactivate;
begin
  DbiDeactivateFilter(Handle, hFilter);
end;

procedure TFilter.Drop;
begin
  DbiDropFilter(Handle, hFilter);
end;

procedure TFilter.MoveValue;
begin
end;

constructor TIntegerFilter.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
end;

function TIntegerFilter.GetValueType : Word;
begin
  Result := fldINT32;
end;

function TIntegerFilter.GetValueSize : Word;
begin
  Result := 4;
end;

procedure TIntegerFilter.MoveValue;
begin
  Move(FValue, CANExp.ValueLiteral, ValueSize);
end;

constructor TFloatFilter.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
end;

function TFloatFilter.GetValueType : Word;
begin
  Result := fldFLOAT;
end;

function TFloatFilter.GetValueSize : Word;
begin
  Result := 8;
end;

procedure TFloatFilter.MoveValue;
begin
  Move(FValue, CANExp.ValueLiteral, ValueSize);
end;

constructor TStringFilter.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
end;

function TStringFilter.GetValueType : Word;
begin
  Result := fldZSTRING;
end;

function TStringFilter.GetValueSize : Word;
begin
  Result := Length(FValue);
end;

procedure TStringFilter.MoveValue;
begin
  Move(FValue[1], CANExp.ValueLiteral, ValueSize);
end;

constructor TDateFilter.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
end;

function TDateFilter.GetValueType : Word;
begin
  Result := fldDATE;
end;

function TDateFilter.GetValueSize : Word;
begin
  Result := 4;
end;

procedure TDateFilter.MoveValue;
begin
  Move(FValue, CANExp.ValueLiteral, ValueSize);
end;

function TDateFilter.GetDate: String16;
begin
  GetDate := DateToStr(FVD);
end;

procedure TDateFilter.SetDate( dtArg: String16 );
begin
FVD:=StrToDate(dtArg);
FValue:=Trunc(fvd);
end;


constructor TTimeFilter.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
end;

function TTimeFilter.GetValueType : Word;
begin
  Result := fldTIME;
end;

function TTimeFilter.GetValueSize : Word;
begin
  Result := 4;
end;

procedure TTimeFilter.MoveValue;
begin
  Move(FValue, CANExp.ValueLiteral, ValueSize);
end;

function TTimeFilter.GetTime: String16;
begin
  GetTime := TimeToStr(FVD);
end;

procedure TTimeFilter.SetTime( dtArg: String16 );
begin
FVD:=StrToTime(dtArg);
FValue:=Round(Frac(fvd)*msecsperday);
end;

constructor TBoolFilter.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
end;

function TBoolFilter.GetValueType : Word;
begin
  Result := fldBool;
end;

function TBoolFilter.GetValueSize : Word;
begin
  Result := 2;
end;

procedure TBoolFilter.MoveValue;
begin
  Fvalue:=FVD;
  Move(FValue, CANExp.ValueLiteral, ValueSize);
end;


constructor TSmallIntFilter.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
end;

function TSmallIntFilter.GetValueType : Word;
begin
  Result := fldINT16;
end;

function TSmallIntFilter.GetValueSize : Word;
begin
  Result := 2;
end;

procedure TSmallIntFilter.MoveValue;
begin
  Move(FValue, CANExp.ValueLiteral, ValueSize);
end;

constructor TTimeStampFilter.Create(AOwner : TComponent);
begin
  fvd:=now;
  Fvalue:= fvd*msecsperday;
  inherited Create(AOwner);

end;

function TTimeStampFilter.GetValueType : Word;
begin
  Result := fldTimeStamp;
end;

function TTimeStampFilter.GetValueSize : Word;
begin
  Result := 8;
end;

procedure TTimeStampFilter.MoveValue;
begin
  Move(FValue, CANExp.ValueLiteral, ValueSize);
end;

function TTimeStampFilter.GetTimeStamp: String32;
begin
  GetTimeStamp := DateTimeToStr(FVD);
end;

procedure TTimeStampFilter.SetTimeStamp( dtArg: String32 );
begin
FVD:=StrToDateTime(dtArg);
FValue:=fvd*msecsperday;
end;


procedure Register;
begin
  RegisterComponents('Div3', [TIntegerFilter, TFloatFilter,
                                TStringFilter,TDateFilter,
                                TTimeFilter,TBoolFilter,
                                TSmallIntFilter,TTimeStampFilter]);
end;

end.
